//
// Copyright (C) 2000 Software Resources International SA
//
// This code is provided on an 'as is' basis.
// Software Resources International SA carries no responsibility for
// the application of this code in any way or form.
//

//Win32 implementation

#include <windows.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "definitions.h"
#include "f11asrt.h"
#include "files11.h"
#include "devio.h"

#ifdef USE_SCSILIB

#include "scsitypes.h"
#include "diskstub.h"
#include "diskerrno.h"
#include "dswin32.h"

#else

// Handle map functions and structures
static HANDLE      *handle_array = NULL;
static int         handle_sz = 0, handle_ff = 0;

static int map_handle(HANDLE h);
static void unmap_handle(int h);
static HANDLE handle_val(int idx);
static void handle_realloc();

static int map_handle(HANDLE h)
{
    int     ret = handle_ff;

    if(handle_ff == handle_sz) handle_realloc();

    handle_array[ret] = h;
    for(; handle_ff < handle_sz && 
            handle_array[handle_ff] == INVALID_HANDLE_VALUE; handle_ff++) 
    {
	;
    }
    return ret;
}

static void unmap_handle(int h)
{
    if(h > handle_sz) return ;
    ASSERT(handle_array != NULL);
    handle_array[h] = INVALID_HANDLE_VALUE;

    if(handle_ff > h) 
	handle_ff = h;

    if(handle_ff == 0) {
        free(handle_array);
        handle_array = NULL;
        handle_sz = 0;
    }
}

static HANDLE handle_val(int idx)
{
    if(idx >= handle_sz) return INVALID_HANDLE_VALUE;
    ASSERT(handle_array != NULL);
    return handle_array[idx];
}

static void handle_realloc() 
{
    int	    i = 0;

    HANDLE  *b = (HANDLE*)malloc((handle_sz + 16) * sizeof(HANDLE));

    for(i = 0; i < 16; i++)
        b[handle_sz + i] = INVALID_HANDLE_VALUE;

    if(handle_array != NULL) { 
        memcpy(b, handle_array, handle_sz * sizeof(HANDLE));
        free(handle_array);
    }

    handle_array = b;
    handle_sz += 16;
}

#endif

int devio_open_dev(const char* name, int mode)
{
#ifdef USE_SCSILIB
    int		h;

    // Open drive
    if((h = DiskOpen_Controller((char*)name, 0)) == -1) {
		TRACE(("Can't open drive '%s'. error %d\n", name, DiskErrNo));
		return INVALID_CHAN;
    }
    return h;

#else
    char	*buf;
    HANDLE      h;
    int		chan = INVALID_CHAN;

    // Make device name
    buf = (char*)malloc(strlen(name) + 6);
    sprintf(buf, "%s", name);

    TRACE(("Device_name '%s'\n", buf));

    h = CreateFile(buf,
		   (mode & DEVIO_RO) ? GENERIC_READ : 
				       GENERIC_READ | GENERIC_WRITE,
                   FILE_SHARE_READ | FILE_SHARE_WRITE,
                   NULL,
                   OPEN_EXISTING,
                   0,
                   0);
    DWORD sys_err = GetLastError();

    free(buf);
    if(h != INVALID_HANDLE_VALUE) {
        chan = map_handle(h);
    }

    return chan;

#endif
}

void devio_close_dev(int dev_chan)
{
#ifdef USE_SCSILIB
    DiskClose(dev_chan);

#else
    CloseHandle(handle_val(dev_chan));
    unmap_handle(dev_chan);

#endif
}

int devio_read_dev(int dev_chan, int blk_number, int blk_count,
	     void* buffer)
{
    LONG    pos = 0;

#ifdef USE_SCSILIB

    int     r_len;

    if(DiskSeek(handle, blk_number * F11_BLOCK_SIZE, 0) == -1) return -1;

    r_len = DiskRead(handle, (char*)buffer, 
		     blk_count * F11_BLOCK_SIZE);
    return r_len == blk_count * F11_BLOCK_SIZE ? r_len : -1;

#else

    HANDLE      h = handle_val(dev_chan);
    DWORD       readlen;
    
    if (SetFilePointer(h, 
                       blk_number * F11_BLOCK_SIZE,
                       NULL,
                       FILE_BEGIN) == 0xFFFFFFFF) 
    {
	DWORD err = GetLastError();
	return -1;
    }
    
    if (ReadFile(h, 
                 buffer, 
                 blk_count * F11_BLOCK_SIZE, 
                 &readlen, 
                 NULL) == FALSE)
    {
	DWORD err = GetLastError();
	return -1;
    }
    if (readlen != (DWORD)(blk_count * F11_BLOCK_SIZE)) 
	return -1;

    return blk_count * F11_BLOCK_SIZE;

#endif
}

int devio_write_dev(int dev_chan, int blk_number, int blk_count,
	     void* buffer)
{
    LONG    pos = 0;

#ifdef USE_SCSILIB

    int     r_len;

    if(DiskSeek(handle,
		blk_number * F11_BLOCK_SIZE,
		0) == -1) 
    {
	return -1;
    }

    r_len = DiskWrite(handle, (char*)buffer, 
		     blk_count * F11_BLOCK_SIZE);
    return r_len == blk_count * F11_BLOCK_SIZE ? r_len : -1;

#else

    HANDLE      h = handle_val(dev_chan);
    DWORD       writelen;
    
    if (SetFilePointer(h, 
                       blk_number * F11_BLOCK_SIZE,
                       NULL,
                       FILE_BEGIN) == 0xFFFFFFFF) 
    {
	DWORD err = GetLastError();
	return -1;
    }
    
    if (WriteFile(h, 
                 buffer, 
                 blk_count * F11_BLOCK_SIZE, 
                 &writelen, 
                 NULL) == FALSE)
    {
	DWORD err = GetLastError();
	return -1;
    }
    if (writelen != (DWORD)(blk_count * F11_BLOCK_SIZE)) 
	return -1;

    return blk_count * F11_BLOCK_SIZE;

#endif
}

int devio_dev_error(void)
{
#ifdef USE_SCSILIB
    return (DWORD)DiskErrNo;
#else
    return -1;
#endif
}

static char error_msg[1024] = {0};

char* devio_dev_error_msg(void)
{
    if(FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM,
		     NULL,
		     GetLastError(),
		     0,
		     error_msg,
		     sizeof(error_msg),
		     NULL) == 0) {
	strcat(error_msg, "Obtaining error information fails.");
    }
	
    return error_msg;
}
