//
// Copyright (C) 2000 Software Resources International SA
//
// This code is provided on an 'as is' basis.
// Software Resources International SA carries no responsibility for
// the application of this code in any way or form.
//

// File-11 Structure level 2

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "definitions.h"
#include "devio.h"
#include "f11asrt.h"
#include "f11liberr.h"
#include "f11file.h"
#include "f11file_seq.h"
#include "f11file_dir.h"
#include "f11vol.h"
#include "f11file_sys.h"
#include "f11lib.h"

// Library data
static f11_volume_acp*	    f11_dev   = NULL;
static f11_directory*	    p_curdir  = NULL;	// current directory
static f11_file_sequential* p_curfile = NULL;	// current file

static char*	f11_fn2string(const ODS_FI2* ida, char *obuf);
static int	get_file_info(const f11_file* f, FileInfoType* di);
static int	get_file_name(const f11_file* f, FileInfoType* di);
static int	set_file_info(f11_file* f, const FileInfoType* di);
static void	f11_unpack_date(ods_word_t* in, DateType* date);

static f11_directory* get_curdir()
{
    return p_curdir;
}

static void change_curdir(f11_directory* curdir) 
{
    ODS_FID	fid;

    if (p_curdir != NULL && 
        p_curdir != f11_dev->p_mfd_dir) 
    {
	f11_dev->qio_deaccess(p_curdir, fid);
	delete p_curdir;
	p_curdir = NULL;
    }
    p_curdir = curdir;
}

//  check lib state macroses
#define CHECK_DEV_OPEN(val) \
if (!f11_dev) { F11_SetError(F11_DEVICE_NOT_OPENED); return val; }

#define CHECK_DEV_NOT_OPEN(val) \
if (f11_dev) { F11_SetError(F11_DEVICE_OPENED); return val; }

#define CHECK_DEV_NOT_RO(val) \
if (f11_dev->mode != DEVIO_RW) { F11_SetError(F11_DEVICE_READ_ONLY); return val; }

#define CHECK_FILE_OPEN(val) \
if (!p_curfile) { F11_SetError(F11_INVALID_HANDLE); return val; }

#define CHECK_FILE_NOT_OPEN(val) \
if (p_curfile) { F11_SetError(F11_FILE_OPENED); return val; }

#define CHECK_DEV_CHAN(h, val)\
if (f11_dev->chan != h) { F11_SetError(F11_INVALID_HANDLE); return val; }

#define CHECK_INPUT_NOTNULL(p,val)\
if (!p) { F11_SetError(F11_INVALID_ARG); return val; }

//------------------------------------------------------------
//  Library Installation
//------------------------------------------------------------
int F11_Install(void)
{
    // Check integrity of the datatypes
    ASSERT(sizeof(byte_t)  == 1);
    ASSERT(sizeof(word_t)  == 2);
    ASSERT(sizeof(dword_t) == 4);
    ASSERT(sizeof(qword_t) == 8);
    ASSERT(sizeof(ODS_FID) == 6);
    ASSERT(sizeof(ACP_FAT) == 32);
    ASSERT(sizeof(ODS_HM2) == 512);
    ASSERT(sizeof(ODS_FH2) == 512);
    ASSERT(sizeof(ODS_SCB) == 512);
    ASSERT(sizeof(ODS_dir_entry) == 8);
    ASSERT(sizeof(ODS_dir_record) == 8);

    return 0;
}

int F11_DeInstall(void)
{
    return 0;
}

//------------------------------------------------------------
//  Device related functions
//------------------------------------------------------------
int F11_Assign(const char* name, int mode)
{
    f11_volume_acp*	dev;
    int		chan;

    CHECK_DEV_NOT_OPEN(INVALID_CHAN);
    CHECK_INPUT_NOTNULL(name, INVALID_CHAN);

    // open device first
    chan = devio_open_dev(name, mode);
    if(chan == INVALID_CHAN) {
	F11_SetError(F11_DRIVE_ERR);
	return INVALID_CHAN;
    }

    dev = new f11_volume_acp;

    // open volume...
    if((dev->f11_assign(chan, mode)) == INVALID_CHAN) {
	devio_close_dev(chan);
	delete dev;
	dev = NULL;
	return INVALID_CHAN;
    }

    // init f11_dev
    change_curdir(dev->p_mfd_dir);
    f11_dev = dev;

    F11_SetError(F11_NO_ERROR);

    return chan;
}

int F11_Deassign(int h)
{
    f11_volume_acp*	dev;
    ODS_FID	fid;

    F11_SetError(F11_NO_ERROR);

    CHECK_DEV_OPEN(-1);
    CHECK_DEV_CHAN(h, -1);

    dev = f11_dev;

    if(p_curfile) {
	dev->qio_deaccess(p_curfile, fid);
	delete p_curfile;
	p_curfile = NULL;
    }

    change_curdir(NULL);

    if (dev->f11_dassgn() == -1)
	return -1;

    delete dev;
    dev = NULL;
    f11_dev = NULL;

    return 0;
}

int F11_GetVolumeInfo(VolInfoType* vi)
{
    CHECK_INPUT_NOTNULL(vi,-1);

    F11_SetError(F11_NO_ERROR);

    if (f11_dev == NULL)
	return -1;

    bzero(vi->volname, 12);
    bzero(vi->ownername, 12);

    //	collect volume info 
    vi->volsize  = f11_dev->p_bitmap_sys->get_content_scb()->scb_l_volsize;
    vi->sectors  = f11_dev->p_bitmap_sys->get_content_scb()->scb_l_sectors;
    vi->cylinder = f11_dev->p_bitmap_sys->get_content_scb()->scb_l_cylinder;
    vi->tracks   = f11_dev->p_bitmap_sys->get_content_scb()->scb_l_tracks;
    vi->rvn      = f11_dev->hblk.hm2_w_rvn;
    vi->cluster  = f11_dev->hblk.hm2_w_cluster;
    vi->maxfiles = f11_dev->hblk.hm2_l_maxfiles;
    vi->extend   = f11_dev->hblk.hm2_w_extend;
    vi->freeblocks = f11_dev->p_bitmap_sys->get_free_clusters() * f11_dev->hblk.hm2_w_cluster;

    strncpy(vi->ownername, (char*)f11_dev->hblk.hm2_t_ownername, 10);
    strncpy(vi->volname, (char*)f11_dev->hblk.hm2_t_volname, 10);

    vi->volowner.Member = (uint8)f11_dev->hblk.hm2_l_volowner.member;
    vi->volowner.Group  = (uint8)f11_dev->hblk.hm2_l_volowner.group;

    vi->fileprot.System = (uint8)f11_dev->hblk.hm2_w_fileprot.system;
    vi->fileprot.Owner  = (uint8)f11_dev->hblk.hm2_w_fileprot.owner;
    vi->fileprot.Group  = (uint8)f11_dev->hblk.hm2_w_fileprot.group;
    vi->fileprot.World  = (uint8)f11_dev->hblk.hm2_w_fileprot.world;

    return 0;
}

//------------------------------------------------------------
//  Directory related functions
//------------------------------------------------------------
int F11_ReadDir(const char* file_spec, FileInfoType* di)
{
    f11_directory*	curdir;
    f11_file*	  	f;
    ODS_FID		fid;
    int			sts;
    word_t		t;
    dword_t		exvbn = 0;

    F11_SetError(F11_NO_ERROR);

    CHECK_DEV_OPEN(-1);
    CHECK_INPUT_NOTNULL(di,-1);

    curdir = get_curdir();
    ASSERT(curdir != NULL);

    f = new f11_file;
    sts = f11_dev->qio_access(f,
			0,		//io_m_create,
			1,		//io_m_access,
			0,		//fib_w_nmctl_fib_v_newver,
			0,		//fib_w_nmctl_fib_v_supersede,
			1,		//fib_w_nmctl_fib_v_wild,
			fid,		//fib_w_fid,
			curdir->get_fhdr()->fh2_w_fid,	//fib_w_did,
			di->WCC,	//fib_l_wcc,
			0,		//fib_b_wsize == 0: do not map the file contents

			0,
			0,
			0,
			0,
			0,
			0,
			exvbn,

			file_spec,	//p2,
			&t,		//p3,
			di->Name	//p4
		);

    if (sts != SS_NORMAL)
    {
		delete f;
		f = NULL;
		return -1; 
    }

    if (!f->is_valid(&curdir->get_fhdr()->fh2_w_fid))
    {
	F11_SetError(F11_INVALID_FH);
	f11_error("Invalid file header");
    }

    // file name
    di->Name[t] = '\0';

    get_file_info(f, di);

    f11_dev->qio_deaccess(f, fid);
    delete f;
    f = NULL;
    return 0;
}

int F11_DirDown(const char* dir_spec)
{
    f11_directory*	curdir;
    f11_directory*	f;
    ODS_FID		fid;
    int			sts;
    dword_t		WCC = 0;
    dword_t		exvbn = 0;

    F11_SetError(F11_NO_ERROR);

    CHECK_DEV_OPEN(-1);
    CHECK_INPUT_NOTNULL(dir_spec, -1);

    curdir = get_curdir();
    ASSERT(curdir != NULL);

    f = new f11_directory;
    sts = f11_dev->qio_access(f,
			0,		//io_m_create,
			1,		//io_m_access,
			0,		//fib_w_nmctl_fib_v_newver,
			0,		//fib_w_nmctl_fib_v_supersede,
			0,		//fib_w_nmctl_fib_v_wild,
			fid,		//fib_w_fid,
			curdir->get_fhdr()->fh2_w_fid,	//fib_w_did,
			WCC,		//fib_l_wcc,
			-1,		//fib_b_wsize == -1: map the entire file

			0,
			0,
			0,
			0,
			0,
			0,
			exvbn,

			dir_spec,	//p2,
			NULL,		//p3,
			NULL		//p4
		);
    if (sts != SS_NORMAL) 
    {
	delete f;
	f = NULL;
	return -1;	// Cannot open
    }

    if(!f->is_valid(&curdir->get_fhdr()->fh2_w_fid))
    {
	f11_dev->qio_deaccess(f, fid);
	delete f;
	f = NULL;
	return -1;	// not directory
    }

    change_curdir(f);

    return 0;
}

int F11_DirUp(void)
{
    f11_directory*	f;
    f11_directory*	curdir;
    ODS_FID*		p_fid;
    int			sts;

    F11_SetError(F11_NO_ERROR);

    CHECK_DEV_OPEN(-1);

    curdir = get_curdir();
    ASSERT(curdir != NULL);

    p_fid = &curdir->get_fhdr()->fh2_w_fid;
    if(p_fid->numb() == FID_MFD_NUMB && p_fid->fid_w_seq == FID_MFD_SEQ) 
    {
	return 0;
    }

    p_fid = &curdir->get_fhdr()->fh2_w_backlink;
    f = new f11_directory;
    sts = f11_dev->acp_sub_access(f, -1, *p_fid);
    if (sts != SS_NORMAL) 
    {
	delete f;
	f = NULL;
	return -1;
    }
    if(!f->is_valid(NULL)) 
    {
	f11_dev->qio_deaccess(f, *p_fid);
	delete f;
	f = NULL;
	return -1;
    }

    change_curdir(f);
    return 0;
}

//------------------------------------------------------------
//  File related functions
//------------------------------------------------------------
int F11_OpenFile(const char* fs, int text_mode)
{
    f11_directory*	curdir;
    f11_file_sequential* f;
    ODS_FID		fid;
    int			sts;
    dword_t		WCC = 0;
    dword_t		exvbn = 0;

    F11_SetError(F11_NO_ERROR);

    CHECK_DEV_OPEN(-1);
    CHECK_FILE_NOT_OPEN(-1);
    CHECK_INPUT_NOTNULL(fs, -1);

    TRACE(("F11_OpenFile: `%s`\n", fs));

    curdir = get_curdir();

    // open file
    f = new f11_file_sequential;
    sts = f11_dev->qio_access(f,
			0,		//io_m_create,
			1,		//io_m_access,
			0,		//fib_w_nmctl_fib_v_newver,
			0,		//fib_w_nmctl_fib_v_supersede,
			0,		//fib_w_nmctl_fib_v_wild,
			fid,		//fib_w_fid,
			curdir->get_fhdr()->fh2_w_fid,	//fib_w_did,
			WCC,		//fib_l_wcc,
			1,		//fib_b_wsize = 1: map only one block

			0,
			0,
			0,
			0,
			0,
			0,
			exvbn,

			fs,		//p2,
			NULL,		//p3,
			NULL		//p4
		);
    if (sts != SS_NORMAL) 
    {
	delete f;
	f = NULL;
	return -1;
    }

    if (!f->is_valid(&curdir->get_fhdr()->fh2_w_fid)) 
    {
	f11_dev->qio_deaccess(f, fid);
	delete f;
	f = NULL;
	return -1;
    }

    f->text_mode = text_mode;
    if (text_mode) {
	TRACE(("OpenFile: file opened in text mode\n"));
    }
    else {
	TRACE(("OpenFile: file opened in binary mode\n"));
    }

    // Store opened file as current
    ASSERT(p_curfile == NULL);
    p_curfile = f;

    return 0;
}

int F11_CreateFile(const char* file_spec, int text_mode, int fsize)
{
    FileInfoType new_file_fi = {0};
    FileInfoType cur_dir_fi  = {0};

    f11_directory*	curdir;
    f11_file_sequential* f;
    ODS_FID		fid;
    int			sts;
    dword_t		WCC = 0;
    dword_t		exvbn = 0;
    dword_t		exts;
    word_t		t;

    F11_SetError(F11_NO_ERROR);

    CHECK_DEV_OPEN(-1);
    CHECK_DEV_NOT_RO(-1);
    CHECK_FILE_NOT_OPEN(-1);
    CHECK_INPUT_NOTNULL(file_spec, -1);

    TRACE(("F11_CreateFile: `%s`\n", file_spec));

    curdir = get_curdir();

    // Inherit some features from the current directory
    get_file_info(curdir, &cur_dir_fi);

    new_file_fi.CreDate = cur_dir_fi.CreDate;
    new_file_fi.RevDate = cur_dir_fi.RevDate;
    new_file_fi.UIC     = cur_dir_fi.UIC;
    new_file_fi.Prot    = cur_dir_fi.Prot;
    new_file_fi.RevDate = cur_dir_fi.RevDate;
    new_file_fi.StrucLev = cur_dir_fi.StrucLev;
    new_file_fi.RMSAttr.FilOrg = 0;
    if (text_mode) {
	new_file_fi.RMSAttr.RecType = 2;
	new_file_fi.RMSAttr.RecAttrib = 0;
	new_file_fi.RMSAttr.MaxRec = 255;
	new_file_fi.RMSAttr.RecSize = 255;
    }
    else {
	new_file_fi.RMSAttr.RecType = 1;
	new_file_fi.RMSAttr.RecAttrib = 0;
	new_file_fi.RMSAttr.MaxRec = 512;
	new_file_fi.RMSAttr.RecSize = 512;
    }

    // create and open file
    f = new f11_file_sequential;
    sts = f11_dev->qio_create(f,
			1,		//io_m_create,
			1,		//io_m_access,
			0,		//io_m_delete,
			fid,		//fib_w_fid,
			curdir->get_fhdr()->fh2_w_fid,	
					//fib_w_did,
			1,		//fib_w_nmctl_fib_v_newver,
			1,		//fib_w_nmctl_fib_v_supersede,
			0,		//fib_w_nmctl_fib_v_wild,
			WCC,		//fib_l_wcc,
			1,		//fib_b_wsize == 1: map the entire file

			0,		//fib_w_exctl_fib_v_extend,
			0,		//fib_w_exctl_fib_v_alcon,
			0,		//fib_w_exctl_fib_v_alconb,
			0,		//fib_w_exctl_fib_v_filcon,
			0,		//fib_w_exctl_fib_v_aldef,
			f11_dev->hblk.hm2_w_extend,
					//fib_l_exsz >= 1 !!!
			exvbn,		//fib_l_exvbn,

			file_spec,	//p2,
			&t,		//p3,
			new_file_fi.Name //p4
		);

    if (sts != SS_NORMAL) 
    {
	delete f;
	f = NULL;
	return -1;
    }

    if (!f->is_valid(&curdir->get_fhdr()->fh2_w_fid)) 
    {
	f11_dev->qio_deaccess(f, fid);
	delete f;
	f = NULL;
	return -1;
    }

    // file name
    new_file_fi.Name[t] = '\0';

    // update file info
    set_file_info(f, &new_file_fi);

    // sync file for the moment
    f->f11_sync();

    f->text_mode = text_mode;

    exts = fsize / f11_dev->hblk.hm2_w_cluster;
    if (fsize % f11_dev->hblk.hm2_w_cluster != 0) exts += 1;
    f->f_size = exts * f11_dev->hblk.hm2_w_cluster;

    if (text_mode) {
	TRACE(("CreateFile: file opened in text mode\n"));
    }

    else {
	TRACE(("CreateFile: file opened in binary mode\n"));
    }

    // store opened file as current
    ASSERT(p_curfile == NULL);
    p_curfile = f;

    return 0;
}

int F11_MkDir(const char* file_spec)
{
    FileInfoType new_dir_fi = {0};
    int mode = 0;   //bin mode

    F11_SetError(F11_NO_ERROR);

    // Create directory file
    f11_directory*	curdir;
    f11_file*		f;
    ODS_FID		fid;
    int			sts;
    word_t		t;
    dword_t		WCC = 0;

    // The directory has initially got some allocated space
    dword_t		exvbn = 0;	
    const char*		fs = (const char*)new_dir_fi.Name;

    CHECK_DEV_OPEN(-1);
    CHECK_DEV_NOT_RO(-1);
    CHECK_FILE_NOT_OPEN(-1);
    CHECK_INPUT_NOTNULL(fs, -1);

    TRACE(("F11_MkDir: `%s`\n", file_spec));

    curdir = get_curdir();

    // Inherit the directory features from the current directory
    get_file_info(curdir, &new_dir_fi);

    // Adjust the file info struct with the filename
    sprintf(new_dir_fi.Name, "%s.DIR;1", file_spec);

    // open file
    f = new f11_file;
    sts = f11_dev->qio_create(f,
			1,		//io_m_create,
			1,		//io_m_access,
			0,		//io_m_delete,
			fid,		//fib_w_fid,
			curdir->get_fhdr()->fh2_w_fid,
					//fib_w_did,
			1,		//fib_w_nmctl_fib_v_newver,
			1,		//fib_w_nmctl_fib_v_supersede,
			0,		//fib_w_nmctl_fib_v_wild,
			WCC,		//fib_l_wcc,
			-1,		//fib_b_wsize == -1: map the entire file

			1,		//fib_w_exctl_fib_v_extend,
			1,		//fib_w_exctl_fib_v_alcon,
			1,		//fib_w_exctl_fib_v_alconb,
			1,		//fib_w_exctl_fib_v_filcon,
			1,		//fib_w_exctl_fib_v_aldef,
			f11_dev->hblk.hm2_w_extend,
					//fib_l_exsz >= 1 !!!
			exvbn,		//fib_l_exvbn,

			fs,		//p2,
			&t,		//p3,
			new_dir_fi.Name //p4
		);

    if (sts != SS_NORMAL) 
    {
	delete f;
	f = NULL;
	return -1;
    }

    if (!f->is_valid(&curdir->get_fhdr()->fh2_w_fid)) 
    {
	f11_dev->qio_deaccess(f, fid);
	delete f;
	f = NULL;
	return -1;
    }

    // file name
    new_dir_fi.Name[t] = '\0';

    // update file info
    set_file_info(f, &new_dir_fi);

    // Set up the initial directory content: 
    // mark up first record as a skip record!
    f->init_empty_dir();

    // Close the newly created directory file
    f11_dev->qio_deaccess(f, fid);

    delete f;
    f = NULL;

    return 0;
}

int F11_GetFileInfo(FileInfoType* fi)
{
    CHECK_DEV_OPEN(-1);
    CHECK_FILE_OPEN(-1);
    CHECK_INPUT_NOTNULL(fi,-1);

    F11_SetError(F11_NO_ERROR);

    get_file_info(p_curfile, fi);
    get_file_name(p_curfile, fi);

    return 0;
}

int F11_Close(void)
{
    ODS_FID	fid;

    CHECK_DEV_OPEN(-1);
    CHECK_FILE_OPEN(-1);

    F11_SetError(F11_NO_ERROR);

    if (p_curfile->record_len != 0)
	p_curfile->flush_record();

    f11_dev->qio_deaccess(p_curfile, fid);
    p_curfile = NULL;

    return 0;
}

int F11_Read(void* buffer, int bytes)
{
    byte_t*	b = (byte_t*)buffer;
    int 	i, j;

    CHECK_DEV_OPEN(-1);
    CHECK_FILE_OPEN(-1);
    CHECK_INPUT_NOTNULL(buffer, -1);
    CHECK_INPUT_NOTNULL(bytes, -1);

    F11_SetError(F11_NO_ERROR);

    TRACE(("read: %d bytes\n", bytes));

    for(i = 0; i < bytes; i++ ) {
	if (p_curfile->text_mode)
	    j = (p_curfile->f11_text_getc)();
	else
	    j = (p_curfile->f11_bin_getc)();

	if(j == -1) {
	    return (F11_GetError() == F11_EOF) ? i : -1;
	}

	b[i] = (byte_t)j;
    }
    return i;
}

int F11_Write(void* buffer, int bytes)
{
    byte_t*	b = (byte_t*)buffer;
    int 	i, j;

    CHECK_DEV_OPEN(-1);
    CHECK_DEV_NOT_RO(-1);
    CHECK_FILE_OPEN(-1);
    CHECK_INPUT_NOTNULL(buffer, -1);
    CHECK_INPUT_NOTNULL(bytes, -1);

    F11_SetError(F11_NO_ERROR);

    TRACE(("write: %d bytes\n", bytes));

    for(i = 0; i < bytes; i++ ) {
	if (p_curfile->text_mode)
	    j = (p_curfile->f11_text_putc)(*b++);
	else
	    j = (p_curfile->f11_bin_putc)(*b++);

	if(j == -1) {
	    break;
	}
    }
    return i;
}

int F11_Delete(const char* file_spec)
{
    f11_directory*	curdir;
    ODS_FID		fid;
    int			sts;
    dword_t		WCC = 0;

    F11_SetError(F11_NO_ERROR);

    CHECK_DEV_OPEN(-1);
    CHECK_DEV_NOT_RO(-1);
    CHECK_INPUT_NOTNULL(file_spec, -1);

    curdir = get_curdir();
    ASSERT(curdir != NULL);

    sts = f11_dev->qio_delete(
			1, 		//io_m_delete,
			1,		//fib_w_nmctl_fib_v_wild,
			fid,		//fib_w_fid,
			curdir->get_fhdr()->fh2_w_fid,
					//fib_w_did,
			WCC,		//fib_l_wcc,
			file_spec,	//p2,
			NULL,		//p3,
			NULL		//p4
		);

    if (sts != SS_NORMAL) 
    {
	return -1;	// Cannot delete
    }

    return 0;
}

//------------------------------------------------------------
//  Error Handling
//------------------------------------------------------------
int F11_GetError(void)
{
    if (f11_dev == NULL)
	return F11_DEVICE_NOT_OPENED;
    else
	return f11_dev->get_error();
}

int F11_GetDevError(void)
{
    if (f11_dev == NULL)
	return 0;
    else
	return f11_dev->get_dev_error();
}

void  F11_SetError(int err)
{
    if (f11_dev != NULL)
	f11_dev->set_error(err);
}

//------------------------------------------------------------
//  file name functions
//------------------------------------------------------------
static char* f11_fn2string(const ODS_FI2* ida, char *obuf)
{
    char        *b = NULL;

    // Copy filename
    bzero(obuf, FH_NAME_LENGTH);
    memcpy(obuf, ida->fi2_t_filename, FH_NAME_SZ);
    if (obuf[FH_NAME_SZ-1] != ' ') {
        memcpy(obuf + FH_NAME_SZ, ida->fi2_t_filenamext, FH_NAMEEX_SZ);
    }

    // Remove pad blanks
    if ((b = strchr(obuf, ' ')) != NULL) {
	*b = '\0';
    }
    return obuf;
}

//------------------------------------------------------------
// Time and Date Conversion
//------------------------------------------------------------
#include <time.h>

// This equals to (1-JAN-1700 - 17-NOV-1858) in seconds
#define DIFF_SEC ((qword_t)0xD1044080) 

static void 
f11_unpack_date(ods_word_t* in, DateType *d)
{
    qword_t	date = (((qword_t)in[0]) + (((qword_t)in[1]) << 16) +
	                (((qword_t)in[2]) << 32) + (((qword_t)in[3]) << 48));
    qword_t	dd   = date / (qword_t) 10000000;  // seconds
    time_t	t;
    struct tm*	stm;

    if(dd < DIFF_SEC) 
	t = 0;
    else 
	t = (time_t)(dd - DIFF_SEC);

    bzero((void*)d, sizeof(DateType));

    TRACE(("f11_unpack_date: %04X%04X%04X%04X\n", (word_t)in[3], (word_t)in[2], (word_t)in[1], (word_t)in[0]));

    // start date (VMS)  	 00:00:00 17 November 1858
    // start date (POSIX time_t) 00:00:00 CUT (Coordinated Universal Time) 
    //				    1 Jan 1970
    stm = localtime(&t);

    d->Century  = (uint8)((stm->tm_year + 1900) / 100);
    d->Year	= (uint8)(stm->tm_year % 100);
    d->Month	= (uint8)(stm->tm_mon + 1);	// [1-12]
    d->Day	= (uint8)stm->tm_mday;		// [1-31]
    d->Hour	= (uint8)stm->tm_hour;		// [0-23]
    d->Minute	= (uint8)stm->tm_min;
    d->Second	= (uint8)stm->tm_sec;		// [0-59]
    d->CentSec	= (uint8)((date / (qword_t)100000) % 100);	
}

//------------------------------------------------------------
//------------------------------------------------------------
static int get_file_info(const f11_file* f, FileInfoType* di)
{
    ODS_FH2*		fh;
    ODS_FI2*		ida;

    // Collect file information
    fh = f->get_fhdr();
    ida = fh->get_id_area();

    // Date & time
    f11_unpack_date(ida->fi2_q_credate, &di->CreDate);
    f11_unpack_date(ida->fi2_q_revdate, &di->RevDate);

    di->UIC.Group  = (uint8)fh->fh2_l_fileowner.group;
    di->UIC.Member = (uint8)fh->fh2_l_fileowner.member;

    di->Prot.System = (uint8)fh->fh2_w_fileprot.system;
    di->Prot.Owner  = (uint8)fh->fh2_w_fileprot.owner;
    di->Prot.Group  = (uint8)fh->fh2_w_fileprot.group;
    di->Prot.World  = (uint8)fh->fh2_w_fileprot.world;

    di->RMSAttr.RecType   = fh->fh2_w_recattr.fat_b_rtype.fat_v_rtype;
    di->RMSAttr.FilOrg 	  = fh->fh2_w_recattr.fat_b_rtype.fat_v_fileorg;
    di->RMSAttr.RecAttrib = fh->fh2_w_recattr.fat_b_rattrib.attr;
    di->RMSAttr.RecSize   = fh->fh2_w_recattr.fat_w_rsize;
    di->RMSAttr.EOFBlk 	  = fh->get_eofblk();
    di->RMSAttr.FFByte    = fh->get_ffbyte();
    di->RMSAttr.MaxRec	  = fh->fh2_w_recattr.fat_w_maxrec;

    di->FileChar[0]	  = fh->fh2_l_filechar.u.fchar[0];
    di->FileChar[1]	  = fh->fh2_l_filechar.u.fchar[1];
    di->FileChar[2]	  = fh->fh2_l_filechar.u.fchar[2];
    di->FileChar[3]	  = fh->fh2_l_filechar.u.fchar[3];

    di->FileNum = fh->fh2_w_fid.numb();
    di->SeqNum  = fh->fh2_w_fid.fid_w_seq;
    di->RVN	= fh->fh2_w_fid.fid_b_rvn;

    di->UsedBlocks  = fh->get_usedblk();
    di->AllocBlocks = fh->get_hiblk();

    TRACE(("get_file_info: done\n"));
    return 0;
}

static int 
get_file_name(const f11_file* f, FileInfoType* di)
{
    ODS_FH2*		fh;
    ODS_FI2*		ida;

    // collect file information
    fh = f->get_fhdr();
    ida = fh->get_id_area();

    // file name
    f11_fn2string(ida, di->Name);

    TRACE(("get_file_name: done\n"));
    return 0;
}

static int 
set_file_info(f11_file* f, const FileInfoType* di)
{
    ODS_FH2*		fh;
    ODS_FI2*		ida;

    // collect file information
    fh = f->get_fhdr();
    ida = fh->get_id_area();

    fh->fh2_l_fileowner.group  = di->UIC.Group;
    fh->fh2_l_fileowner.member = di->UIC.Member;

    fh->fh2_w_fileprot.system = di->Prot.System;
    fh->fh2_w_fileprot.owner  = di->Prot.Owner;
    fh->fh2_w_fileprot.group  = di->Prot.Group;
    fh->fh2_w_fileprot.world  = di->Prot.World;

    fh->fh2_w_recattr.fat_b_rtype.fat_v_rtype   = di->RMSAttr.RecType;
    fh->fh2_w_recattr.fat_b_rtype.fat_v_fileorg = di->RMSAttr.FilOrg;
    fh->fh2_w_recattr.fat_b_rattrib.attr        = di->RMSAttr.RecAttrib;
    fh->fh2_w_recattr.fat_w_rsize               = di->RMSAttr.RecSize;
    fh->fh2_w_recattr.fat_w_maxrec              = di->RMSAttr.MaxRec;

    fh->fh2_l_filechar.u.fchar[0] = di->FileChar[0];
    fh->fh2_l_filechar.u.fchar[1] = di->FileChar[1];
    fh->fh2_l_filechar.u.fchar[2] = di->FileChar[2];
    fh->fh2_l_filechar.u.fchar[3] = di->FileChar[3];

    f->flag_header_modified = 1;

    return 0;
}

int F11_SetTraceOn(void)
{
#if defined(_DEBUG)
    f11_enable_trace();
    return 0;
#else
    return -1;
#endif
}

int F11_SetTraceOff(void)
{
#if defined(_DEBUG)
    f11_disable_trace();
    return 0;
#else
    return -1;
#endif
}
