//
// Copyright (C) 2000 Software Resources International SA
//
// This code is provided on an 'as is' basis.
// Software Resources International SA carries no responsibility for
// the application of this code in any way or form.
//

// Files-11 Structure level 2

#include <string.h>
#include <stdio.h>

#include "definitions.h"
#include "devio.h"
#include "f11asrt.h"
#include "files11.h"

//------------------------------------------------------------
//  Local Function Declarations
//------------------------------------------------------------
static word_t calc_checksum(ods_word_t *block, int len = 255, int start = 0);

//------------------------------------------------------------
//  File indentifier functions
//------------------------------------------------------------
ODS_FID::ODS_FID(dword_t numb, word_t seq, byte_t rvn)
{
    ASSERT(!(numb & 0xFF000000));

    fid_w_num = numb & 0x0FFFF;
    fid_b_rvn = rvn;
    fid_w_seq = seq;
    fid_b_nmx = (numb >> 16) & 0x0FF;
}

//------------------------------------------------------------
//  File name (Ident Area) functions
//------------------------------------------------------------
char* ODS_FI2::fn2string(char *obuf) const
{
    char    *b = NULL;

    // Copy filename
    bzero(obuf, FH_NAME_LENGTH);
    memcpy(obuf, this->fi2_t_filename, FH_NAME_SZ);
    if(obuf[FH_NAME_SZ - 1] != ' ')
        memcpy(obuf + FH_NAME_SZ, this->fi2_t_filenamext, FH_NAMEEX_SZ);

    // Remove pad blanks
    if((b = strchr(obuf, ' ')) != NULL) 
	*b = '\0';

    return obuf;
}

void ODS_FI2::string2fn(const char *obuf)
{
    const char        *b = obuf;

    // Copy filename
    unsigned int len = sprintf((char*)this->fi2_t_filename, "%-*.*s", FH_NAME_SZ, FH_NAME_SZ, b);
    ASSERT(len <= FH_NAME_LENGTH);

    if (len < strlen(obuf)) {
	b = b + len;
	int len = sprintf((char*)this->fi2_t_filenamext, "%-*.*s", FH_NAMEEX_SZ-1, FH_NAMEEX_SZ-1, b);
    }
}

//------------------------------------------------------------
//  File header functions
//------------------------------------------------------------
int ODS_FH2::has_valid_checksum() const
{
    word_t  chksumm;

    chksumm = calc_checksum((ods_word_t*)this);
    if (chksumm != fh2_w_checksum) 
	return 0;

    return 1;
}

void ODS_FH2::update_checksum()
{
    word_t  chksumm;

    chksumm = calc_checksum((ods_word_t*)this);
    fh2_w_checksum = chksumm;
}

int ODS_FH2::is_file_header(const ODS_FID* fid) const
{
    // Identify File Header Block
    if (!has_valid_checksum()) 
	return 0;

    // Check other predictable values
    if(fh2_b_idoffset > fh2_b_mpoffset 	  ||
       fh2_b_mpoffset > fh2_b_acoffset 	  ||
       fh2_b_acoffset > fh2_b_rsoffset	  ||
       fh2_w_struclev.ver == 0		  ||
       fh2_w_struclev.lev != DEF_STRUCT_LEV ||
       fh2_w_fid.fid_w_num == FID_INVALID_NUMB ||
       fh2_b_map_inuse > (fh2_b_acoffset - fh2_b_mpoffset))
    {
	return 0;
    }

    // fid must be equal
    if(!(fh2_w_fid == *fid))
	return 0;
	
    return 1;
}

int ODS_FH2::is_file_deleted() const
{
    if (fh2_l_filechar.u.s.fch_v_markdel &&
        fh2_w_fid.fid_w_num == 0	&&
        fh2_w_fid.fid_b_rvn == 0	&&
        fh2_w_fid.fid_b_nmx == 0)
        return 1;

    return 0;
}

int  ODS_FH2::is_file_contig() const
{
    return (fh2_l_filechar.u.s.fch_v_contig);
}

int 
ODS_FH2::is_file_directory() const
{
    return (fh2_l_filechar.u.s.fch_v_directory &&
	    is_file_contig() &&
	    fh2_w_recattr.fat_b_rattrib.rattr.fat_m_nospan);
}

dword_t ODS_FH2::get_hiblk() const
{
    return ((dword_t)fh2_w_recattr.fat_l_hiblk.fat_w_hiblkh << 16) | 
	    (dword_t)fh2_w_recattr.fat_l_hiblk.fat_w_hiblkl;
}

dword_t ODS_FH2::get_eofblk() const
{
    // When the end-of-file position corresponds to a block
    // boundary, by convention FAT$L_EFBLK contains the end-of-file VBN plus 1,
    // and FAT$W_FFBYTE contains 0.
    // So when HIBLK == 0, i.e there is no space allocated for the
    // file, EOFBLK must be equal to 1 and FFBYTE == 0.
    dword_t eofblk = ((dword_t)fh2_w_recattr.fat_l_efblk.fat_w_efblkh << 16) | 
		      (dword_t)fh2_w_recattr.fat_l_efblk.fat_w_efblkl;
    ASSERT(eofblk != 0);
    return eofblk;
}

word_t ODS_FH2::get_ffbyte() const
{
    return (word_t)fh2_w_recattr.fat_w_ffbyte;
}

// Return eofblk with correction
dword_t ODS_FH2::get_usedblk() const
{
    // When the end-of-file position corresponds to a block
    // boundary, by convention FAT$L_EFBLK contains the end-of-file VBN plus 1,
    // and FAT$W_FFBYTE contains 0.
    // Note: USEDBLK == 0 only when EOFBLK == 1 and FFBYTE == 0.
    // It is (only) possible while no space is allocated for the file
    // (i.e. HIBLK == 0).
    dword_t usedblk = get_eofblk() - ((get_ffbyte() == 0) ? 1 : 0);
    ASSERT(usedblk >= 0);
    return usedblk;
}

void ODS_FH2::set_hiblk(dword_t hiblk)
{
    fh2_w_recattr.fat_l_hiblk.fat_w_hiblkh = (word_t)(hiblk >> 16);
    fh2_w_recattr.fat_l_hiblk.fat_w_hiblkl = (word_t)hiblk & 0xFFFF;
}

void ODS_FH2::set_eofblk(dword_t eofblk)
{
    fh2_w_recattr.fat_l_efblk.fat_w_efblkh = (word_t)(eofblk >> 16);
    fh2_w_recattr.fat_l_efblk.fat_w_efblkl = (word_t)eofblk & 0xFFFF;
}

void ODS_FH2::set_ffbyte(word_t ffbyte)
{
    fh2_w_recattr.fat_w_ffbyte = ffbyte;
}

dword_t ODS_FH2::get_file_length() const
{
    // When the end-of-file position corresponds to a block
    // boundary, by convention efblk contains the end-of-file vbn plus 1,
    // and ffbyte contains 0.
    dword_t	eofblk = get_eofblk();
    word_t	ffbyte = get_ffbyte();

    return (eofblk - 1) * F11_BLOCK_SIZE + ffbyte;
}

void ODS_FH2::set_file_length(dword_t len)
{
    word_t	ffbyte = len % F11_BLOCK_SIZE;
    dword_t	eofblk = len / F11_BLOCK_SIZE;

    // When the end-of-file position corresponds to a block
    // boundary, by convention efblk contains the end-of-file vbn plus 1,
    // and ffbyte contains 0.
    eofblk += 1;

    set_eofblk(eofblk);
    set_ffbyte(ffbyte);
}

// Return block count mapped by all the retrieval pointers 
// in the map area
dword_t	ODS_FH2::get_rp_blk_count() const
{
    dword_t	blk_count = 0;
    ODS_FM2  	*rp;

    // Go through all used retrieval pointers
    // untill FH2$B_MAP_INUSE instances passed
    rp = get_mp_area();
    rp_iterator next_rp(rp, fh2_b_map_inuse);
    while ((rp = next_rp++) != NULL)
    {
	// Handle placement header found? just skip it...
	if(rp->rp_isfmt0()) {
	    continue;
	}
	blk_count += rp->rp_getcount();
    }

    return blk_count;
}

//------------------------------------------------------------
//  Directory functions
//------------------------------------------------------------
dword_t ODS_dir_record::get_entry_count() const
{
    if(!dir_w_size) return 0;
    return ((dword_t)dir_w_size + 2 - 6 - dir_b_namecount - (dir_b_namecount & 1)) / 
		sizeof(ODS_dir_entry);
}

ODS_dir_entry* ODS_dir_record::get_entry(dword_t numb) const
{

    ODS_dir_entry* ret = (ODS_dir_entry*)((byte_t*)this + 6 + 
			dir_b_namecount + (dir_b_namecount & 1)) + numb;
    return ret;
}

//------------------------------------------------------------
//  Retrieval pointer functions
//------------------------------------------------------------
lbn_t ODS_FM2::rp_getlbn() const
{
    // Check for proper alignment
    ASSERT(!((size_t)this & 1));

    switch(ret_ptr0.fm2_v_format) {
      case FM2_C_PLACEMENT:
	ASSERT(0);
	break;

      case FM2_C_FORMAT1:
	return (lbn_t)ret_ptr1.fm2_w_lowlbn[0] |
	      ((lbn_t)ret_ptr1.fm2_w_lowlbn[1] << 8) |
	      ((lbn_t)ret_ptr1.fm2_v_highlbn << 16);

      case FM2_C_FORMAT2:
	return (lbn_t)ret_ptr2.fm2_l_lbn2[0] |
	      ((lbn_t)ret_ptr2.fm2_l_lbn2[1] << 8) |
	      ((lbn_t)ret_ptr2.fm2_l_lbn2[2] << 16) |
	      ((lbn_t)ret_ptr2.fm2_l_lbn2[3] << 24);

      case FM2_C_FORMAT3:
	return (lbn_t)ret_ptr3.fm2_l_lbn3[0] |
	      ((lbn_t)ret_ptr3.fm2_l_lbn3[1] << 8) |
	      ((lbn_t)ret_ptr3.fm2_l_lbn3[2] << 16) |
	      ((lbn_t)ret_ptr3.fm2_l_lbn3[3] << 24);
    }
    return INVALID_LBN;
}

dword_t ODS_FM2::rp_getwords() const
{
    // Check for proper alignment
    ASSERT(!((size_t)this & 1));

    switch(ret_ptr0.fm2_v_format) {
      case FM2_C_PLACEMENT: 
	  return sizeof(ret_ptr0) / 2;
      case FM2_C_FORMAT1: 
	  return sizeof(ret_ptr1) / 2;
      case FM2_C_FORMAT2: 
	  return sizeof(ret_ptr2) / 2;
      case FM2_C_FORMAT3: 
	  return sizeof(ret_ptr3) / 2;
    }
    ASSERT(0);
    return sizeof(ret_ptr0) / 2;
}

dword_t	ODS_FM2::rp_getcount() const
{
    // Check for proper alignment
    ASSERT(!((size_t)this & 1));

    switch(ret_ptr0.fm2_v_format) {
      case FM2_C_PLACEMENT: 
	ASSERT(0);
	break;

      case FM2_C_FORMAT1: 
	return (dword_t)ret_ptr1.fm2_v_count1 + 1;

      case FM2_C_FORMAT2:
	return (dword_t)((dword_t)ret_ptr2.fm2_v_count2_1 |
			((dword_t)ret_ptr2.fm2_v_count2_2 << 8)) + 1;

      case FM2_C_FORMAT3:
	return ((dword_t)ret_ptr3.fm2_w_lowcount[0] |
	       ((dword_t)ret_ptr3.fm2_w_lowcount[1] << 8) |
	       ((dword_t)ret_ptr3.fm2_v_count2_1 << 16) |
	       ((dword_t)ret_ptr3.fm2_v_count2_2 << 24)) + 1;
    }
    return 0;
}

int ODS_FM2::rp_isfmt0() const
{
    return (ret_ptr0.fm2_v_format == FM2_C_PLACEMENT);
}

void ODS_FM2::rp_setlbn(lbn_t lbn)
{
    switch(ret_ptr0.fm2_v_format) {
      case FM2_C_PLACEMENT:
	ASSERT(0);
	break;

      case FM2_C_FORMAT1:
	ret_ptr1.fm2_w_lowlbn[0] = (byte_t)lbn;
	ret_ptr1.fm2_w_lowlbn[1] = (byte_t)(lbn >> 8);
	ret_ptr1.fm2_v_highlbn = (byte_t)((lbn >> 16) & 0x3F);
	break;

      case FM2_C_FORMAT2:
	ret_ptr2.fm2_l_lbn2[0] = (byte_t)lbn;
	ret_ptr2.fm2_l_lbn2[1] = (byte_t)(lbn >> 8);
	ret_ptr2.fm2_l_lbn2[2] = (byte_t)(lbn >> 16);
	ret_ptr2.fm2_l_lbn2[3] = (byte_t)(lbn >> 24);
	break;

      case FM2_C_FORMAT3:
	ret_ptr3.fm2_l_lbn3[0] = (byte_t)lbn;
	ret_ptr3.fm2_l_lbn3[1] = (byte_t)(lbn >> 8);
	ret_ptr3.fm2_l_lbn3[2] = (byte_t)(lbn >> 16);
	ret_ptr3.fm2_l_lbn3[3] = (byte_t)(lbn >> 24);
	break;
    }
}

void ODS_FM2::rp_setcount(dword_t count)
{
    switch(ret_ptr0.fm2_v_format) {
      case FM2_C_PLACEMENT: 
	ASSERT(0);
	break;

      case FM2_C_FORMAT1: 
	ret_ptr1.fm2_v_count1  = count - 1;
	break;

      case FM2_C_FORMAT2:
	ret_ptr2.fm2_v_count2_1 = (byte_t)(count-1);
	ret_ptr2.fm2_v_count2_2 = (byte_t)((count >> 8) & 0x003F);
	break;

      case FM2_C_FORMAT3:
	ret_ptr3.fm2_w_lowcount[0] = (byte_t)(count - 1);
	ret_ptr3.fm2_w_lowcount[1] = (byte_t)(count >> 8);
	ret_ptr3.fm2_v_count2_1 = (byte_t)(count >> 16);
	ret_ptr3.fm2_v_count2_2 = (byte_t)((count >> 24) & 0x003F);
	break;
    }
}

void ODS_FM2::rp_init(int format, lbn_t lbn, dword_t count)
{
    ASSERT(format == FM2_C_PLACEMENT ||
	   format == FM2_C_FORMAT1 ||
	   format == FM2_C_FORMAT2 ||
	   format == FM2_C_FORMAT3);

    ret_ptr0.fm2_v_format = format;
    rp_setlbn(lbn);
    rp_setcount(count);
}

//------------------------------------------------------------
//  Home block functions   
//------------------------------------------------------------
int ODS_HM2::has_valid_checksum() const
{
    word_t	chksumm;

    chksumm = calc_checksum((ods_word_t*)this, CHECKSUMM1_WORDS);
    if (chksumm != hm2_w_checksum1) 
	return 0;

    chksumm = calc_checksum((ods_word_t*)this);
    if (chksumm != hm2_w_checksum2) 
	return 0;

    return 1;
}

void ODS_HM2::update_checksum()
{
    word_t	chksumm;

    chksumm = calc_checksum((ods_word_t*)this, CHECKSUMM1_WORDS);
    hm2_w_checksum1 = chksumm;

    chksumm = calc_checksum((ods_word_t*)this);
    hm2_w_checksum2 = chksumm;
}

int ODS_HM2::is_valid() const
{
    // Identify Home Block
    if (!has_valid_checksum()) 
	return 0;

    // Check other predictable values
    if (strncmp((char*)hm2_t_format, FORMAT_STRING, strlen(FORMAT_STRING)) != 0 ||
        hm2_l_homelbn	    == 0 ||
        hm2_l_alhomelbn	    == 0 ||
        hm2_l_altidxlbn	    == 0 ||
        hm2_w_struclev.ver  == 0 || 
        hm2_w_struclev.lev  != DEF_STRUCT_LEV ||
        hm2_w_cluster	    == 0 ||
        hm2_w_homevbn	    == 0 ||
        hm2_l_ibmaplbn	    == 0 ||
        hm2_l_maxfiles	    >  F11_MAX_FILES ||
	hm2_l_maxfiles	    <= hm2_w_resfiles ||
        hm2_w_ibmapsize	    == 0 ||
        hm2_w_resfiles	    <  F11_MIN_RESFILES ||
	hm2_w_devtype	    != 0
	) 
    {
	return 0;
    }
    return 1;
}

//------------------------------------------------------------
//  Storage Control Block functions   
//------------------------------------------------------------
int ODS_SCB::has_valid_checksum() const
{
    word_t	chksumm;

    chksumm = calc_checksum((ods_word_t*)this);
    if (chksumm != scb_w_checksum) 
	return 0;

    return 1;
}

void ODS_SCB::update_checksum()
{
    word_t	chksumm;

    chksumm = calc_checksum((ods_word_t*)this);
    scb_w_checksum = chksumm;
}

int ODS_SCB::is_valid() const
{
    // Identify Storage Control Block
    if (!has_valid_checksum()) 
	return 0;

    // Check other predictable values
    if (scb_w_struclev.ver  == 0 || 
        scb_w_struclev.lev  != DEF_STRUCT_LEV ||
        scb_w_cluster	    == 0 ||
        scb_l_volsize	    == 0
	) 
    {
	return 0;
    }
    return 1;
}

dword_t	ODS_SCB::get_bitmap_size()
{
    ASSERT((scb_l_volsize % scb_w_cluster) != 0);
    ASSERT(((scb_l_volsize / scb_w_cluster) % 8) != 0);

    return (scb_l_volsize / scb_w_cluster);
}

//------------------------------------------------------------
//  Local Function Definitions
//------------------------------------------------------------

// calc_checksum: produces one of block checksum values...
static word_t calc_checksum(ods_word_t *block, int len, int start)
{
    register int count = len;
    register word_t result = 0;
    register ods_word_t *ptr = block + start;

    do {
        result += *ptr++;
    } while (--count > 0);

    return result;
}
