//
// Copyright (C) 2000 Software Resources International SA
//
// This code is provided on an 'as is' basis.
// Software Resources International SA carries no responsibility for
// the application of this code in any way or form.
//

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>

#ifdef __linux__
#include <limits.h>
#endif

#ifdef unix
#define stricmp strcasecmp
#endif

#include "definitions.h"
#include "fileio.h"
#include "devio.h"
#include "f11asrt.h"
#include "ft11.h"

static int    ft11_dev  = INVALID_CHAN;
static char   ft11_dev_name[PATH_MAX + NAME_MAX + 1] = {0};
static char   ft11_current_dir[1024] = "Not connected.";
static char   file_name[256];
static char   all_files[] = "*.*;*";

// Default mode is binary
static int    ft11_text_mode = 0;	

static char* f11_errors[] = {
    "No error.",
    "Drive error.",
    "Device isn't opened.",
    "Device already opened.",
    "Invalid handle.",
    "File already opened.",
    "Invalid arument.",
    "Home block isn't found or invalid.",
    "Invalid file index.",
    "Invalid sequence number.",
    "Invalid file header.",
    "Invalid file header back link.",
    "Volume sets doesn't supported.",
    "EOF reached.",
    "File not found.",
    "File isn't directory.",
    "Already top directory.",
    "No free file index available.",
    "Device is read-only.",
    "File already exists.",
    "Invalid text_file.",
    "Error reading file."
};

#define F11_MAX_ERROR_MSG_LENGTH 1024
static char f11_error_msg[F11_MAX_ERROR_MSG_LENGTH];

static char* F11_GetErrorString(void)
{
    if (F11_GetError() >= sizeof(f11_errors) / sizeof(char*)) {
        sprintf(f11_error_msg, "Unknown error code '%d'", F11_GetError());
        return f11_error_msg;
    }

    if (F11_GetError() != F11_DRIVE_ERR) 
	return f11_errors[F11_GetError()];
    
    // Make full error message
    sprintf(f11_error_msg, "%s %s", f11_errors[F11_GetError()], 
	    devio_dev_error_msg());
    return f11_error_msg;
}

static char *pdate_month[] = {
    "JAN", "FEB", "MAR", "APR", "MAY", "JUN", 
    "JUL", "AUG", "SEP", "OCT", "NOV", "DEC"
}; 

static void pdate(const DateType *d)
{
    ASSERT(d->Century >= 18 && d->Year < 100);
    ASSERT(d->Day <= 31 && d->Day);
    ASSERT(d->Month <= 12 && d->Month);
    ASSERT(d->Minute < 60 && d->Second < 60 && d->Hour < 24);
    ASSERT(d->CentSec < 100);

    printf("%2d-%s-%04d %02d:%02d:%02d.%02d", (int)d->Day, pdate_month[d->Month - 1],
           (((int)(d->Century)) * 100 + d->Year),
           (int)d->Hour, (int)d->Minute, (int)d->Second, (int)d->CentSec);
}

static char pprot_str[] = { 'R', 'W', 'E', 'D' };

static void pprot(const ProtType *p)
{
    uint8   *w = (uint8 *)p;
    int	    i, j;

    for (i = 0; i < 4; i++, w++) {
        for (j = 0; j < 4; j++) {
            if (!(*w & (1 << j))) 
		printf("%c", pprot_str[j]);
        }
        if (i != 3) 
	    printf(",");
    }
}

static void ft11_dir_down(const char *dir)
{
    ASSERT(strlen(ft11_current_dir));

    ft11_current_dir[strlen(ft11_current_dir) - 1] = '\0';
    strcat(ft11_current_dir, ".");
    strcat(ft11_current_dir, dir);
    strcat(ft11_current_dir, "]");
}

static void ft11_dir_up()
{
    char *b = ft11_current_dir + strlen(ft11_current_dir);
    if(strchr(ft11_current_dir, '.') == NULL) return;
    else {
	for (; *b != '.' && b > ft11_current_dir; b--) ;
	ASSERT(b != ft11_current_dir);
	*b++ = ']'; *b = '\0';
    }
}

void ft11_help_cmd(int arg_numb, char* args[])
{
    printf("Commands available are:\n");
    printf("  op*en               by*e                q*uit               ex*it	    \n");
    printf("  lo                  cl*ose              dis*connect         ls	    \n");
    printf("  cd                  cdu*p               mk*dir              pwd	    \n");
    printf("  get                 put                 re*cv               vi	    \n");
    printf("  sh*ow_dev           show_dev/full       show_def            dir	    \n");
    printf("  dir/d*ate           dir/f*ile_id        dir/p*rot           dir/s*ize \n");
    printf("  del                 delete              asc*ii              bi*n	    \n");
    printf("  mo*de               deb*ug              lcd                 h*elp	    \n");
    printf("  st*atus             ?			          !	\n");
    printf("Abbreviations to (*) are allowed\n");
    printf("\n");
}

void ft11_open_cmd(int arg_numb, char* args[])
{
    int		dev;
    int		rw_mode = DEVIO_RW;
    
    // Process arguments
    if (!arg_numb) {
	f11_error("Invalid command arguments\n "
	      "Usage: open <devname> [rw|ro]");
	return;
    }

    if (arg_numb >= 2)
    {
        if (stricmp(args[1], "rw") == 0) 
	    rw_mode = DEVIO_RW;
        else if(stricmp(args[1], "ro") == 0) 
	    rw_mode = DEVIO_RO;
        else
        {
            printf("Invalid mode '%s'.\n", args[1]);
            return;
        }
    }

    if (ft11_dev != INVALID_CHAN)
    {
        printf("Already connected to '%s', use disconnect first.\n", ft11_dev_name);
        return;
    }

    //  Open device
    if ((dev = F11_Assign(args[0], rw_mode)) == INVALID_CHAN) {
	f11_error("Cannot open volume '%s'. %s", 
		  args[0], F11_GetErrorString());
	return;
    }

    strncpy(ft11_dev_name, args[0], sizeof(ft11_dev_name));
    if (rw_mode == DEVIO_RO)
	printf("Volume '%s' connected for read only access.\n", ft11_dev_name);
    else
	printf("Volume '%s' connected for read/write access.\n", ft11_dev_name);

    ft11_dev = dev;
    strcpy(ft11_current_dir, "[000000]");
}

void ft11_bye_cmd(int arg_numb, char* args[])
{
    F11_Deassign(ft11_dev);
    sprintf(ft11_current_dir, "Not connected.");
    ft11_dev  = INVALID_CHAN; 

    printf("Goodbye.\n");
    cmd_done = 1;
}

void ft11_close_cmd(int arg_numb, char* args[])
{
    if (ft11_dev == INVALID_CHAN) {
	printf("Not connected.\n");
	return;
    }

    F11_Deassign(ft11_dev);
    
    printf("Volume '%s' disconnected.\n", ft11_dev_name);
    sprintf(ft11_current_dir, "Not connected.");
    
    ft11_dev = INVALID_CHAN;
}

void ft11_bin_cmd(int arg_numb, char* args[])
{
    ft11_text_mode = 0;
    printf("Type set to I.\n");
}

void ft11_ascii_cmd(int arg_numb, char* args[])
{
    ft11_text_mode = 1;
    printf("Type set to A.\n");
}

void ft11_mode_cmd(int arg_numb, char* args[])
{
    //  Process arguments
    if (!arg_numb) {
	f11_error("Invalid command arguments\n "
	      "Usage: mode [ascii|bin]");
	return;
    }

    if (arg_numb >= 1)
    {
        if (stricmp(args[0], "bin") == 0) 
	    ft11_text_mode = 0;
        else if(stricmp(args[0], "ascii") == 0) 
	    ft11_text_mode = 1;
        else
        {
            printf("Invalid mode '%s'.\n", args[0]);
            return;
        }
    }
}

void ft11_debug_cmd(int arg_numb, char* args[])
{
    // Process arguments
    if (!arg_numb) {
	f11_error("Invalid command arguments\n "
	      "Usage: debug [on|off]");
	return;
    }

    if (arg_numb >= 1)
    {
        if (stricmp(args[0], "on") == 0) {
	    if (F11_SetTraceOn() != 0)
		printf("This version does not contain debug support.\n");
	    else
		printf("Trace is ON.\n");
	}
        else if(stricmp(args[0], "off") == 0) {
	    if (F11_SetTraceOff() != 0)
		printf("This version does not contain debug support.\n");
	    else
		printf("Trace is OFF.\n");
	}
        else {
            printf("Invalid debug mode '%s'.\n", args[0]);
            return;
        }
    }
}

void ft11_ls_cmd(int arg_numb, char* args[])
{
    FileInfoType	di;
    dword_t		i = 0, fsize = 0;
    const char*		file_spec;

    bzero((void*)&di, sizeof(FileInfoType));

    printf("Directory %s\n", ft11_current_dir);

    //  Process arguments
    if (arg_numb == 0) {
	file_spec = all_files;
    } else {
	file_spec = args[0];
    }

    while (F11_ReadDir(file_spec, &di) == 0) {
	printf("%s\n\tOwner: [%d,%d]", di.Name, di.UIC.Member, di.UIC.Group);
        fsize = di.RMSAttr.EOFBlk ?
            (di.RMSAttr.EOFBlk - 1) * F11_BLOCK_SIZE + di.RMSAttr.FFByte : 0;
        printf("\tSize: %d (%d bytes)\tProtection: ", di.RMSAttr.EOFBlk, fsize);
        pprot(&di.Prot);
        printf("\n\tCreated: "); pdate(&di.CreDate);
        printf("\tRevised: "); pdate(&di.RevDate);
        printf("\n");
        fflush(stdout);
	if (F11_GetError() != F11_NO_ERROR) 
		f11_error("%s", F11_GetErrorString());
	i++;
    }

    if (F11_GetError() != F11_FILE_NOT_FOUND) {
	f11_error("Can't list directory. %s", F11_GetErrorString());
    }

    printf("Total of %d files\n", i);
}

void ft11_dir_cmd(int arg_numb, char* args[])
{
    FileInfoType	di;
    dword_t		i = 0, fsize = 0;
    const char*		file_spec;

    bzero((void*)&di, sizeof(FileInfoType));

    //  Process arguments
    if (arg_numb == 0) {
	file_spec = all_files;
    } else {
	file_spec = args[0];
    }

    if (F11_ReadDir(file_spec, &di) == -1)
    {
	if (F11_GetError() == F11_FILE_NOT_FOUND) {
	    f11_info("%%DIRECT-W-NOFILES, no files found");
	    return;
	} else {
	    f11_error("%s", F11_GetErrorString());
	    return;
	}
    }

    printf("\n");
    printf("Directory DKA100:%s\n", ft11_current_dir);
    printf("\n");

    do {
	printf("%s\n", di.Name);
	if (F11_GetError() != F11_NO_ERROR) 
	    f11_error("%s", F11_GetErrorString());
	i++;
    } while (F11_ReadDir(file_spec, &di) == 0);

    printf("\n");
    printf("Total of %d files.\n", i);
}

void ft11_dir_date_cmd(int arg_numb, char* args[])
{
    FileInfoType	di;
    dword_t		i = 0, fsize = 0;
    const char*		file_spec;

    bzero((void*)&di, sizeof(FileInfoType));

    //  Process arguments
    if (arg_numb == 0) {
	file_spec = all_files;
    } else {
	file_spec = args[0];
    }

    if (F11_ReadDir(file_spec, &di) == -1)
    {
	if (F11_GetError() == F11_FILE_NOT_FOUND) {
	    f11_info("%%DIRECT-W-NOFILES, no files found");
	    return;
	} else {
	    f11_error("%s", F11_GetErrorString());
	    return;
	}
    }

    printf("\n");
    printf("Directory DKA100:%s\n", ft11_current_dir);
    printf("\n");

    do {
	int n = printf("%-19s", di.Name);
	if (n > 19)
	    printf("\n                     ");
	else
	    printf("  ");

	pdate(&di.CreDate);
	printf("\n");
	if (F11_GetError() != F11_NO_ERROR) 
		f11_error("%s", F11_GetErrorString());
	i++;
    } while (F11_ReadDir(file_spec, &di) == 0);

    printf("\n");
    printf("Total of %d files.\n", i);
}

void ft11_dir_fileid_cmd(int arg_numb, char* args[])
{
    FileInfoType	di;
    dword_t		i = 0, fsize = 0;
    const char*		file_spec;

    bzero((void*)&di, sizeof(FileInfoType));

    //  Process arguments
    if (arg_numb == 0) {
	file_spec = all_files;
    } else {
	file_spec = args[0];
    }

    if (F11_ReadDir(file_spec, &di) == -1)
    {
	if (F11_GetError() == F11_FILE_NOT_FOUND) {
	    f11_info("%%DIRECT-W-NOFILES, no files found");
	    return;
	} else {
	    f11_error("%s", F11_GetErrorString());
	    return;
	}
    }

    printf("\n");
    printf("Directory DKA100:%s\n", ft11_current_dir);
    printf("\n");

    do
    {
	int n = printf("%-19s", di.Name);
	if (n > 19)
	    printf("\n                     ");
	else
	    printf("  ");
	printf("(%lu,%u,%u)", di.FileNum, di.SeqNum, di.RVN);
	printf("\n");
	if (F11_GetError() != F11_NO_ERROR) 
		f11_error("%s", F11_GetErrorString());
	i++;
    } while (F11_ReadDir(file_spec, &di) == 0);

    printf("\n");
    printf("Total of %d files.\n", i);
}

void ft11_dir_prot_cmd(int arg_numb, char* args[])
{
    FileInfoType	di;
    dword_t		i = 0, fsize = 0;
    const char*		file_spec;

    bzero((void*)&di, sizeof(FileInfoType));

    //  Process arguments
    if (arg_numb == 0) {
	file_spec = all_files;
    } else {
	file_spec = args[0];
    }

    if (F11_ReadDir(file_spec, &di) == -1)
    {
	if (F11_GetError() == F11_FILE_NOT_FOUND) {
	    f11_info("%%DIRECT-W-NOFILES, no files found");
	    return;
	} else {
	    f11_error("%s", F11_GetErrorString());
	    return;
	}
    }

    printf("\n");
    printf("Directory DKA100:%s\n", ft11_current_dir);
    printf("\n");

    do {
	int n = printf("%-19s", di.Name);
	if (n > 19)
	    printf("\n                     ");
	else
	    printf("  ");
	printf("(");
	pprot(&di.Prot);
	printf(")");
	printf("\n");
	if (F11_GetError() != F11_NO_ERROR) 
		f11_error("%s", F11_GetErrorString());
	i++;
    } while (F11_ReadDir(file_spec, &di) == 0);

    printf("\n");
    printf("Total of %d files.\n", i);
}

void ft11_dir_size_cmd(int arg_numb, char* args[])
{
    FileInfoType	di;
    dword_t		i = 0, fsize = 0;
    dword_t		total_blk_used = 0, total_blk_alloc = 0;
    const char*		file_spec;

    bzero((void*)&di, sizeof(FileInfoType));

    //  Process arguments
    if (arg_numb == 0) {
	file_spec = all_files;
    } else {
	file_spec = args[0];
    }

    if (F11_ReadDir(file_spec, &di) == -1)
    {
	if (F11_GetError() == F11_FILE_NOT_FOUND) {
	    f11_info("%%DIRECT-W-NOFILES, no files found");
	    return;
	} else {
	    f11_error("%s", F11_GetErrorString());
	    return;
	}
    }

    printf("\n");
    printf("Directory DKA100:%s\n", ft11_current_dir);
    printf("\n");

    do {
	int n = printf("%-19s", di.Name);
	if (n > 19)
	    printf("\n                     ");
	else
	    printf("  ");
	printf("%7lu/%-7lu", di.UsedBlocks, di.AllocBlocks);
	printf("\n");
	if (F11_GetError() != F11_NO_ERROR) 
		f11_error("%s", F11_GetErrorString());
	i++;
	total_blk_used  += di.UsedBlocks; 
	total_blk_alloc += di.AllocBlocks; 
    } while (F11_ReadDir(file_spec, &di) == 0);

    printf("\n");
    printf("Total of %d files, %lu/%lu blocks.\n", i, total_blk_used, total_blk_alloc);
}

void ft11_cd_cmd(int arg_numb, char* args[])
{
    char	*b;
    int		err;
    char        dir_name[FH_NAME_LENGTH], dir_full[FH_NAME_LENGTH];

    //  Process arguments
    if (arg_numb != 1) {
	f11_error("Invalid number of parameters\n"
	    " Usage: cd NAME[[.DIR];VER]|..");
	return;
    }
    if (strlen(args[0]) > FH_NAME_LENGTH - 4) {
        f11_error("Directory name too long");
        return;
    }
    for (b = args[0]; *b != 0; b++) 
	*b = toupper(*b);

    bzero(dir_name, FH_NAME_LENGTH);
    bzero(dir_full, FH_NAME_LENGTH);

    b = strchr(args[0], '.');
    if (!b) {
        strcpy(dir_name, args[0]);
        sprintf(dir_full, "%s.DIR;1", args[0]);
    } else {
        strncpy(dir_name, args[0], b - args[0]);
        strcpy(dir_full, args[0]);
    }

    if(strstr(dir_full, "000000") != 0) return;
    
    if (strcmp(args[0], "..") == 0) {
	err = F11_DirUp();
        if (err != -1) ft11_dir_up();        
    } else {
	err = F11_DirDown(dir_full);
        if (err != -1) ft11_dir_down(dir_name);
    }

    if (err == -1) {
	f11_error("Can't open directory '%s'. %s", dir_full, F11_GetErrorString());
	return;    
    }

    printf("Working directory is %s\n", ft11_current_dir);
}

void ft11_mkdir_cmd(int arg_numb, char* args[])
{
    char	*b;
    int		err;
    char        dir_name[FH_NAME_LENGTH], dir_full[FH_NAME_LENGTH];

    //  Process arguments
    if (arg_numb != 1) {
	f11_error("Invalid number of parameters\n"
	    " Usage: mkdir NAME");
	return;
    }
    if (strlen(args[0]) > FH_NAME_LENGTH - 4) {
        f11_error("Directory name too long");
        return;
    }
    for (b = args[0]; *b != 0; b++) 
	*b = toupper(*b);

    bzero(dir_name, FH_NAME_LENGTH);
    bzero(dir_full, FH_NAME_LENGTH);

    b = strchr(args[0], '.');
    if (!b) {
        strcpy(dir_name, args[0]);
    } else {
        strncpy(dir_name, args[0], b - args[0]);
    }
    
    err = F11_MkDir(dir_name);

    if (err == -1) {
	f11_error("Cannot create directory '%s'. %s", dir_name, F11_GetErrorString());
	return;
    }
    printf("Directory %s created\n", dir_name);
}

void ft11_cdup_cmd(int arg_numb, char* args[])
{
    char  cdup_str[] = "..";
    char* cdup_args[] = {cdup_str};
    ft11_cd_cmd(1, cdup_args);
}

void ft11_vi_cmd(int arg_numb, char* args[])
{
    VolInfoType	vi;
    if (F11_GetVolumeInfo(&vi) == -1) {
	f11_error("Cannot obtain volume info. %s", F11_GetErrorString());
	return;
    }

    //  Print volume info
    printf("volume image file is '%s' .\n", ft11_dev_name);
    printf("volume label is '%10s'.\n", vi.volname);
    printf("volume ownername is '%10s'.\n", vi.ownername);
    printf("volume size is '%d'.\n", vi.volsize);
    printf("volume sectors is '%d'.\n", vi.sectors);
    printf("volume cylinder is '%d'.\n", vi.cylinder);
    printf("volume tracks is '%d'.\n", vi.tracks);
    printf("volume rvn is '%d'.\n", vi.rvn);
    printf("volume cluster is '%d'.\n", vi.cluster);
    printf("volume maxfiles is '%d'.\n", vi.maxfiles);
    printf("volume extend is '%d'.\n", vi.extend);
    printf("volume freeblocks is '%d'.\n", vi.freeblocks);
}

void ft11_show_dev_cmd(int arg_numb, char* args[])
{
    bool	    mounted;
    VolInfoType	    vi;

    mounted = (F11_GetVolumeInfo(&vi) == -1) ? false : true;

    //  Print volume info
    printf("\n");
    printf("Device                 Device           Error    Volume         Free  Trans Mnt\n");
    printf(" Name                  Status           Count     Label        Blocks Count Cnt\n");
    if (mounted) {
	printf("DKA100:                Mounted            N/A  %12.12s"   "%10lu   N/A   1\n", vi.volname, vi.freeblocks);
    } else {
	printf("DKA100:                Online             N/A                                  \n");
    }
}

void ft11_show_dev_full_cmd(int arg_numb, char* args[])
{
    bool	    mounted;
    VolInfoType	    vi;

    mounted = (F11_GetVolumeInfo(&vi) == -1) ? false : true;

    //  Print full volume info
    printf("\n");
    printf("Disk        DKA100:, device type   N/A, is online, ");
    if (mounted) {
	printf("mounted, ");
    }
    printf("file-oriented\n");
    printf("    device, error logging is disabled.\n");
    printf("\n");
    printf("    Error count                  N/A    Operations completed                N/A\n");
    printf("    Owner process        \"F11ODS2FTU\"    Owner UIC                  %12.12s\n",		vi.ownername);
    printf("    Owner process ID             N/A    Dev Prot                            N/A\n");
    printf("    Reference count              N/A    Default buffer size                 512\n");
    printf("    Total blocks          %10lu    Sectors per track            %10lu\n",			vi.volsize, vi.sectors);
    printf("    Total cylinders       %10lu    Tracks per cylinder          %10lu\n",			vi.cylinder, vi.tracks);
    printf("\n");
    if (mounted) {
	printf("    Volume label      \"%12.12s\"    Relative volume number           %5u\n",		vi.volname, vi.rvn);
	printf("    Cluster size               %5u    Transaction count                   N/A\n",		vi.cluster);
	printf("    Free blocks           %10lu    Maximum files allowed        %10lu\n",			vi.freeblocks, vi.maxfiles);
	printf("    Extend quantity            %5u    Mount count                           1\n",		vi.extend);
	printf("    Mount status              System    Cache name                          N/A\n");
	printf("    Extent cache size            N/A    Maximum blocks in extent cache      N/A\n");
	printf("    File ID cache size           N/A    Blocks currently in extent cache    N/A\n");
	printf("    Quota cache size             N/A    Maximum buffers in FCP cache        N/A\n");
	printf("    Volume owner UIC    %12.12s    Vol Prot    S:,O:,G:,W:\n",				vi.ownername);
	printf("\n");
	printf("  Volume Status: subject to mount verification, file high-water marking, write-\n");
	printf("      back caching disabled.\n");
	printf("\n");
    }
}

void ft11_pwd_cmd(int arg_numb, char* args[])
{
    printf("'%s' is current directory.\n", ft11_current_dir);
}

void ft11_show_def_cmd(int arg_numb, char* args[])
{
    printf("  DKA100:%s\n", ft11_current_dir);
}

#define FT11_GETBUFFER_LEN F11_BLOCK_SIZE
#define FT11_PUTBUFFER_LEN F11_BLOCK_SIZE

void ft11_get_cmd(int arg_numb, char* args[])
{
    HANDLE		dest_file;
    int			src_file;
    dword_t		i;
    int			bytes_read;
    byte_t		buf[FT11_GETBUFFER_LEN];
    char		*file_spec;
	int			err = 0;

    //  Options
    if (!arg_numb) {
	f11_error("Invalid number of parameters\n "
	      "Usage: get <FILENAME>.<TYPE>[;<VERSION>]");
	return;
    }
    file_spec = args[0];

    //  Make destination file name having no version
    for (i = 0; i < strlen(file_spec) && file_spec[i] != ';' ; i++)
	file_name[i] = file_spec[i];	
    file_name[i] = '\0';

    //  Convert the file name to upper case
    for (i = 0; i < strlen(file_spec); i++)
	file_spec[i] = toupper(file_spec[i]);

    //  Open source file
    if ((src_file = F11_OpenFile(file_spec, ft11_text_mode)) == -1) {
	f11_error("Can't open file '%s'. %s", file_spec, F11_GetErrorString());
	return;
    }

    //  Open destination file
    dest_file = fileio_open_file(file_name, FILEIO_RW);
    if (dest_file == INVALID_HANDLE_VALUE) {
	f11_error("Can't open destination file '%s'. %s", 
		    file_name, devio_dev_error_msg());
	F11_Close();
	return;
    }

    printf("Copying DKA100:%s to '%s' ...\n", file_spec, file_name);

    //	Read file 
    bytes_read = FT11_GETBUFFER_LEN;

    while (F11_GetError() != F11_EOF) {

		bytes_read = F11_Read(buf, FT11_GETBUFFER_LEN);
		
		if (bytes_read == -1 && F11_GetError() == F11_NO_ERROR) {
			err = 0;
			break;
		}

		if (bytes_read == -1 && F11_GetError() != F11_NO_ERROR) {
		    f11_error("Error reading source file '%s'. %s", 
				file_spec, F11_GetErrorString());
			err = 1;
			break;
		}

		TRACE(("was read: %d bytes\n", bytes_read));

		if (fileio_write_file(dest_file, buf, bytes_read) == -1 ) 
		{
			f11_error("Error writing destination file '%s'. %s",
				file_name, devio_dev_error_msg());
			err = 1;
			break;
		}
    }

    if (!err) printf("Transfer complete.\n");

    fileio_close_file(dest_file);
    F11_Close();
}

void ft11_put_cmd(int arg_numb, char* args[])
{
    HANDLE		source_file;
    int			dst_file;
    dword_t		i, fsize, size;
    int			bytes_read;
    byte_t		buf[FT11_PUTBUFFER_LEN];
    char		*file_spec;
	int			err = 0;

    VolInfoType	    vi;

    //  Options
    if (F11_GetVolumeInfo(&vi) == -1) {
	f11_error("Volume not mounted\n ");
	return;
    }

    if (!arg_numb) {
	f11_error("Invalid number of parameters\n "
	      "Usage: get <FILENAME>.<TYPE>[;<VERSION>]");
	return;
    }
    file_spec = args[0];

    //  Make the source file name without verson
    for (i = 0; i < strlen(file_spec) && file_spec[i] != ';' ; i++)
	file_name[i] = file_spec[i];
    file_name[i] = '\0';

    //  Convert the file name to upper case
    for (i = 0; i < strlen(file_spec); i++) 
	file_spec[i] = toupper(file_spec[i]); 
    
    //  Open source file
    source_file = fileio_open_file(file_name, FILEIO_RO);
    if (source_file == INVALID_HANDLE_VALUE) {
	f11_error("Can't open source file '%s'. %s", 
	      file_name, devio_dev_error_msg());
	return;
    }

    size = fileio_file_size(file_name);
    if(size == (dword_t)-1) return;
    fsize = size / FT11_PUTBUFFER_LEN;
    if (size % FT11_PUTBUFFER_LEN != 0) fsize +=1;
    if (fsize == 0) fsize = 1;

    if (fsize >= vi.freeblocks) {
	f11_error("Source file size (%d blocks) exceed free space (%d blocks)", fsize, vi.freeblocks);
	fileio_close_file(source_file);
	return;
    }

    //  Open destination file
    if ((dst_file = F11_CreateFile(file_spec, ft11_text_mode, fsize)) == -1) {
	f11_error("Can't create destination file '%s'. %s", file_spec, F11_GetErrorString());
	fileio_close_file(source_file);
	return;
    }

    printf("Copying '%s' - %lu bytes to DKA100:%s ...\n", file_name, size, file_spec);

    //	Read file
    bytes_read = FT11_PUTBUFFER_LEN;

    while (bytes_read) {

		bytes_read = fileio_read_file(source_file, buf, FT11_PUTBUFFER_LEN);
		if (bytes_read == -1) {
		    f11_error("Error reading source file '%s'. %s", 
				file_name, devio_dev_error_msg());
			err = 1;
			break;
		}

		TRACE(("to be written: %d bytes\n", bytes_read));

		int bbb = F11_Write(buf, bytes_read);

		if (bytes_read && bbb != bytes_read) {
			f11_error("Error writing destination file '%s'. %s",
				file_spec, F11_GetErrorString());
			err = 1;
			break;
		}
    }
    
    if (!err) printf("Transfer complete.\n");

    fileio_close_file(source_file);
    F11_Close();
}

void ft11_lcd_cmd(int arg_numb, char* args[])
{
    char 	dir[PATH_MAX + 1];

    if (!arg_numb) {
	f11_error("Invalid number of parameters\n "
		  "Usage: lcd <DIRNAME>");
	return ;
    }

    //  Change directory
    if (fileio_file_chdir(args[0]) == -1) {
	f11_error("'%s': File not found. %s", 
		  args[0], devio_dev_error_msg());
	return;
    }

    printf("Local directory now '%s'.\n", 
	   fileio_file_getwd(dir, PATH_MAX));
}

//------------------------------------------------------------
//
void ft11_del_cmd(int arg_numb, char* args[])
{
    char*	file_spec;

    //  Options
    if (!arg_numb) {
	f11_error("Invalid number of parameters\n "
		  "Usage: del <FILENAME>.<TYPE>;<VERSION>");
	return;
    }
    file_spec = args[0];

    //  Convert the file name to upper case
    for (dword_t i = 0; i < strlen(file_spec); i++) 
	file_spec[i] = toupper(file_spec[i]); 

    if(strstr(file_spec, ".DIR") != NULL) {
	f11_error("Directory cannot be deleted\n" );
	return;
    }

    if(strstr(file_spec, ";") == NULL) {
	f11_error("File version should be entered\n" );
	return;
    }

    if(F11_Delete(file_spec) == -1) 
    {
	printf("Cannot delete file '%s'. %s\n", 
		    file_spec, devio_dev_error_msg());
	return;
    }

    printf("File '%s' deleted.\n", file_spec);
}        

